from outset._auxlib.make_radial_gradient_ import make_radial_gradient


def test_gradient_shape():
    """Test if the output array has the correct shape."""
    gradient = make_radial_gradient()
    assert len(gradient.shape) == 2, "Gradient should be two dimensional"


def test_gradient_value_range():
    """Test if the gradient values are within the range 0 to 1."""
    gradient = make_radial_gradient()
    assert (
        gradient.min() >= 0
    ), "Minimum value should be greater than or equal to 0"
    assert (
        gradient.max() <= 1
    ), "Maximum value should be less than or equal to 1"



import typing


def get_vertices_extent(
    vertices: typing.Iterable[typing.Tuple[float, float]]
) -> typing.Tuple[float, float, float, float]:
    """Get the extent of a set of vertices.

    Parameters
    ----------
    vertices : iterable of (float, float)
        The vertices to find the extent of.

    Returns
    -------
    tuple of float
        The extent of the vertices in the form (xmin, xmax, ymin, ymax).

    """
    x, y = zip(*vertices)
    return (min(x), max(x), min(y), max(y))



import typing

from matplotlib import axes as mpl_axes
import numpy as np


def compose_callout_leader(
    frame_xlim: typing.Tuple[float, float],
    frame_ylim: typing.Tuple[float, float],
    ax: mpl_axes.Axes,
    stretch: float,
    stretch_unit: typing.Literal["axes", "figure", "inches", "inchesfrom"],
) -> typing.List[typing.Tuple[float, float]]:
    """Decide shape, size, and position of callout leader triangle.

    Outer leader vertex placement is calculated relative to the upper right
    frame vertex.

    Uses reflected right triangle similar to the proportions of the rectangle
    to pick the upper points for the trapezoid, relative to the upper right
    corner of the rectangle. The trapezoid is then constructed to include
    the upper left corner of the rectangle and the lower right corner of the rectangle.

    The distance of the upper vertices of the zoom annotation from the
    rectangle are controlled by scaling the similar triangles. Scaling is
    adjusted by relative size of the rectangle within the axis viewport,
    with smaller rectangles given proportionally larger stretch factors to
    prevent overlap of decorations applied to the upper vertices with the
    rectangle.

    This can be tweaked through the "stretch" parameter, with higher stretch
    factors resulting in upper vertices further from the rectangle.

    Note that a portion of the trapezoid falls behind the rectangle.

    Limits should be provided in ascending order.

    Parameters
    ----------
    frame_xlim : typing.Tuple[float, float]
        The x-axis limits (min, max) of the rectangle.
    frame_ylim : typing.Tuple[float, float]
        The y-axis limits (min, max) of the rectangle.
    ax : mpl_axes.Axes
        The axes object where the callout leader will be drawn.
    stretch : float
        The stretch factor for the zoom effect.
    stretch_unit : {"axes", "figure", "inches", "inchesfrom"}, default "axes"
        How should stretch be specified?

    Returns
    -------
    list
        A list of tuples representing the sequence of vertices of the trapezoid, arranged to minimize the perimeter.
    """
    if any(float.__gt__(*map(float, lim)) for lim in [frame_xlim, frame_ylim]):
        raise ValueError("Limits must be provided as (min, max).")

    if stretch < 0.0:
        raise ValueError("Leader stretch must be non-negative.")

    ax_width, ax_height = np.ptp(ax.get_xlim()), np.ptp(ax.get_ylim())
    frame_width, frame_height = np.ptp(frame_xlim), np.ptp(frame_ylim)
    (
        (frame_lower_left, frame_upper_right),
        (frame_upper_left, frame_lower_right),
    ) = zip(frame_xlim, frame_ylim), zip(frame_xlim, reversed(frame_ylim))

    if frame_width:
        theta = np.arctan(
            frame_height
            / frame_width
            * np.ptp(ax.get_xlim())
            / np.ptp(ax.get_ylim())
        )
    elif frame_height:
        theta = np.pi / 2
    else:
        theta = 0

    stretch_y, stretch_x = np.sin(theta) * stretch, np.cos(theta) * stretch
    if stretch_unit == "axes":
        height, width = stretch_y * ax_height, stretch_x * ax_width
    elif stretch_unit == "figure":
        fig_width, fig_height = ax.figure.get_size_inches()
        pixel_width, pixel_height = ax.transData.transform(
            [fig_width * stretch_x, fig_height * stretch_y],
        )
        # Convert display coordinates back to data coordinates
        height, width = ax.transData.inverted().transform(
            [(pixel_width, pixel_height)],
        )[0]
    elif stretch_unit == "inches":
        axwidth_inches = ax.get_position().width * ax.figure.get_figwidth()
        axheight_inches = ax.get_position().height * ax.figure.get_figheight()
        stretch_x_axfrac = stretch_x / axwidth_inches
        stretch_y_axfrac = stretch_y / axheight_inches
        height, width = (
            stretch_y_axfrac * ax_height,
            stretch_x_axfrac * ax_width,
        )
    elif stretch_unit == "inchesfrom":
        axwidth_inches = ax.get_position().width * ax.figure.get_figwidth()
        axheight_inches = ax.get_position().height * ax.figure.get_figheight()
        stretch_x_axfrac = stretch_x / axwidth_inches
        stretch_y_axfrac = stretch_y / axheight_inches
        width, height = (
            stretch_x_axfrac * ax_width,
            stretch_y_axfrac * ax_height,
        )
        minimum_lower_left = np.array(frame_lower_left) + np.array(
            [width, height],
        )
        width, height = np.maximum(
            minimum_lower_left - np.array(frame_upper_right),
            np.zeros(2),
        )
    else:
        raise ValueError(
            f"stretch_unit must be 'axes', 'figure', 'inches', or 'inchesfrom' "
            "not {stretch_unit}",
        )
    offsets = np.array([width, height])

    leader_outer_vertex = tuple(np.array(frame_upper_right) + offsets)

    leader_vertices = [
        frame_upper_left,
        frame_upper_right,
        frame_lower_right,
        leader_outer_vertex,
    ]

    return leader_vertices



import itertools as it
import typing

import pandas as pd
from matplotlib import pyplot as plt
import seaborn as sns
from typing import Optional, Any, Dict


def regplot(
    data: pd.DataFrame,
    *,
    x: str,
    y: str,
    hue: Optional[str] = None,
    hue_order: Optional[Any] = None,
    ax: Optional[plt.Axes] = None,
    **kwargs: Dict[str, Any],
) -> plt.Axes:
    """Plot regressions with seaborn's regplot on a pandas DataFrame.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame containing the data to plot.
    x : str
        The name of the column in `data` to use for the x-axis values.
    y : str
        The name of the column in `data` to use for the y-axis values.
    hue : Optional[str], default None
        The name of the column in `data` to use for color encoding.
    hue_order : Optional[Any], default None
        The order to plot the `hue` levels, if `hue` is not None.
    ax : Optional[plt.Axes], default None
        The matplotlib Axes object to draw the plot onto, if provided.
    **kwargs
        Additional keyword arguments forward to seaborn's regplot.

    Returns
    -------
    plt.Axes
        The matplotlib Axes containing the plot.

    Notes
    -----
    This function extends seaborn's regplot functionality by adding support for
    hue-based grouping and customizing plot aesthetics.
    """
    data = data.copy()
    if ax is None:
        ax = plt.gca()
    palette = kwargs.pop("palette", sns.color_palette())
    if hue is None:
        hue = "_dummy_hue"
        data[hue] = 0

    if hue_order is None:
        hue_order = sorted(data[hue].unique())

    data["_hue_order"] = data[hue].map(hue_order.index)
    data.sort_values("_hue_order")

    for color, (_value, group) in zip(
        it.cycle(palette),
        data.groupby(hue, sort=False),
    ):
        sns.regplot(
            data=group,
            x=x,
            y=y,
            ax=ax,
            **{
                "color": color,
                **kwargs,
            },
        )
    return ax



from matplotlib import axes as mpl_axes
import numpy as np


def calc_aspect(ax: mpl_axes.Axes) -> float:
    """Calculate the aspect ratio of the axes.

    If physical is True, the aspect ratio is calculated based on the size ratio
    between unit values unit on each axis after rendering transforms. If
    physical is False, the aspect ratio is based on the axis viewport limits.
    """
    xlim, ylim = ax.get_xlim(), ax.get_ylim()
    res = ax._get_aspect_ratio()
    # @MAM WTF, matplotlib is mutating axis limits???
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)
    assert (xlim, ylim) == (ax.get_xlim(), ax.get_ylim())
    return res



import itertools as it

from matplotlib import pyplot as plt
from matplotlib import patches as mpl_patches
import numpy as np
import pytest
import seaborn as sns

from outset import mark as otst_mark
from outset._auxlib.lighten_color_ import lighten_color


@pytest.mark.integration
def test_make_glyph_menu():
    markers = {
        "MarkAlphabeticalBadges\nstart='a'": otst_mark.MarkAlphabeticalBadges(
            "a"
        ),
        "MarkAlphabeticalBadges\nstart='A'": otst_mark.MarkAlphabeticalBadges(
            "A"
        ),
        "MarkArrow": otst_mark.MarkArrow(),
        "MarkInlaidAsterisk": otst_mark.MarkInlaidAsterisk(),
        "MarkMagnifyingGlass": otst_mark.MarkMagnifyingGlass(),
        "MarkNumericalBadges": otst_mark.MarkNumericalBadges(),
        "MarkRomanBadges\nupper=False": otst_mark.MarkRomanBadges(upper=False),
        "MarkRomanBadges\nupper=True": otst_mark.MarkRomanBadges(upper=True),
    }

    fig, axs = plt.subplots(ncols=len(markers), figsize=(len(markers), 1))
    for (name, mark), color, ax in zip(
        markers.items(), it.cycle(sns.color_palette()), np.array(axs).flat
    ):
        ax.set_axis_off()
        ax.add_patch(
            mpl_patches.Rectangle(
                (0, 0),
                1,
                1,
                edgecolor="none",
                facecolor=lighten_color(color),
            )
        )
        mark(0.25, 0.75, ax=ax, color=color)
        mark(0.6, 0.4, ax=ax, color=color)
        ax.set_title(name, color=color, loc="left", rotation=60)
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)

    fig.tight_layout()

    # Save and print the path of the output image
    outpath = "/tmp/test_make_glyph_menu.png"
    plt.savefig(outpath, bbox_inches="tight")
    print(f"saved graphic to {outpath}")



import pandas as pd

from outset._auxlib.robust_groupby_ import robust_groupby

df = pd.DataFrame({"A": [1, 2, 1, 2], "B": [5, 6, 7, 8], "C": [9, 10, 11, 12]})


def test_groupby_single_column():
    grouped = robust_groupby(df, by=["A"])
    assert list(grouped.groups.keys()) == [1, 2]


def test_groupby_multiple_columns():
    grouped = robust_groupby(df, by=["A", "B"])
    assert list(grouped.groups.keys()) == [(1, 5), (1, 7), (2, 6), (2, 8)]


def test_groupby_with_none():
    grouped = robust_groupby(df, by=[None, "B"])
    assert list(grouped.groups.keys()) == [5, 6, 7, 8]


def test_groupby_empty_list():
    grouped = robust_groupby(df, by=[])
    assert len(grouped) == 1 and grouped[0][0] is None and grouped[0][1] is df


def test_groupby_only_none_list():
    grouped = robust_groupby(df, by=[None])
    assert len(grouped) == 1 and grouped[0][0] is None and grouped[0][1] is df
